/* ----------------------------------------------------------------------- */
/* -                     Astronomical Dome Control                       - */
/* -                             INDI Driver                             - */
/* ----------------------------------------------------------------------- */
/* Copyright (c) 2009 John Kielkopf                                        */
/* kielkopf@louisville.edu                                                 */
/*                                                                         */
/* This file is part of XmTel.                                             */
/*                                                                         */
/* XmTel is free software: you can redistribute it and/or modify           */
/* it under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation, either version 3 of the License, or       */
/* (at your option) any later version.                                     */
/*                                                                         */
/* XmTel is distributed in the hope that it will be useful,                */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* Credits: INDI components included here were derived from software       */
/* Copyright (C) 2003 Elwood C. Downey and originally distributed under    */
/* the terms of the GNU Lesser General Public License version 2.1.         */             
/*                                                                         */
/* Date: November 27, 2009                                                 */
/* Version: 1.3.7                                                          */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "indidevapi.h"
#include "dome.h"
#include "protocol.h"


/* Dependencies */

/* indidrivermain.c */
/* eventloop.c */
/* liblilxml */
/* astro.h */
/* protocol.c */
/* algorithms.c */


/* Dome parameters */

double site_longitude;      /* Preferred naming convention */
double site_latitude;       /* Preferred naming convention */
double SiteLongitude;       /* Needed by algorithms */
double SiteLatitude;        /* Needed by algorithms */
double dec_offset;          /* Center of dome to dec axis */   
double dec_height;          /* Height of dec axis at polar axis */
double dec_length;          /* Length from polar to optical axes */
double dome_radius;         /* Radius of dome */
int    telmount = GEM;      /* One of GEM (2) , EQFORK (1) , ALTAZ (0) */
double upperlimit = 360.;   /* Upper azimuth limit for dome rotation */
double lowerlimit = 0.;     /* Lower azimuth limit for dome rotation */

/* Internal control flags */

int domeflag;
int domestatus;
int domegoto = FALSE;
int telpoll = TELPOLLOFF;
int upper_shutter_status = 0;
int lower_shutter_status = 0;

/* Dome azimuth */

double domeaz;
double dometargetaz;


/* Hardware-dependent external routines defined in the protocols */
/* The server calls these routines to control the telescope */
 
/* Dome and shutter operations */

extern void ControlDome(int domecmd);
extern int  DomeToAzimuth(double dometargetaz); 
extern void SolveDomeAzimuth(double ra, double dec, double *az, int nloops);
extern void GetDome(double *domeaz, int *status);
extern void GetShutter(int *upperstatus, int *lowerstatus);
extern void GetTel(double *ra, double *dec);
extern void SetDomeLimits(double lower, double upper);

/* ----------------------------------- */
/* ---- End of external routines  ---- */
/* ----------------------------------- */

  
/* INDI prototypes and variables */


/* Functions for indiserver */

static void initdome();
static void dome (void *p);
void read_config(void); 
            

/*                                            */
/* INDI number                                */
/*                                            */
/* Index name                                 */
/* Short description                          */
/* GUI display format                         */
/* Range, ignore if min == max                */
/* Step size, ignore if step == 0             */
/* Current value                              */
/* Pointer to parent vector property          */
/* Opt: handy place to hang helper info       */
/*                                            */
/*                                            */
/* INDI number property                       */
/*                                            */
/* Device name                                */
/* Property name                              */
/* Short description                          */
/* GUI grouping hint                          */
/* Client accessibility hint                  */
/* Current max time to change, secs           */
/* Current property state                     */
/* Numbers comprising this vector             */
/* Dimension of np[]                          */
/* ISO 8601 timestamp of this event           */
/* Opt: handy place to hang helper info       */


/* Dome azimuth */

static INumber indi_dome_azimuth[] = 
{ 
  {"degrees", "Degrees",   "%.0f",   0., 360., 1., 0.},
};

static INumberVectorProperty indi_dome_azimuth_property = 
{ 
  MYDEV, 
  "azimuth", 
  "Azimuth", 
  "", 
  IP_RO, 
  0, 
  IPS_IDLE, 
  indi_dome_azimuth, 
  NARRAY(indi_dome_azimuth) 
};

/* Dome target azimuth */

static INumber indi_dome_target[] = 
{ 
  {"degrees", "Degrees",   "%.0f",   0., 360., 1., 0.},
};

static INumberVectorProperty indi_dome_target_property = 
{ 
  MYDEV, 
  "target", 
  "Target", 
  "", 
  IP_RW, 
  0, 
  IPS_IDLE, 
  indi_dome_target, 
  NARRAY(indi_dome_target) 
};


/* INDI switches                                   */
/*                                                 */
/* Index name                                      */
/* Switch label                                    */
/* Switch state                                    */
/* Opt: Pointer to parent                          */
/* Opt: Handy place to hang helper info            */
/*                                                 */
/* Device name                                     */
/* Property name                                   */
/* Short description will appear on a client GUI   */
/* GUI grouping hint                               */
/* Client accessibility hint                       */
/* Switch behavior hint                            */
/* Current max time to change, secs                */
/* Current property state                          */
/* Switches comprising this vector                 */
/* Dimension of sp[] (with help from NARRAY)       */
/* ISO 8601 timestamp of this event                */
/* Opt: handy place to hang helper info            */


/* Read telescope azimuth switch */

static ISwitch indi_dome_telaz_state[] = 
{ 
  {"off", "Off", ISS_ON},
  {"monitor", "Monitor", ISS_OFF},
  {"track", "Track", ISS_OFF},
};

/* Read telescope vector property */

static ISwitchVectorProperty indi_dome_telaz_state_property = 
{
  MYDEV, 
  "telescope", 
  "Telescope ", 
  "", 
  IP_RW,
  ISR_1OFMANY, 
  0, 
  IPS_IDLE, 
  indi_dome_telaz_state, 
  NARRAY(indi_dome_telaz_state)
};


/* Go to switch */

static ISwitch indi_dome_goto_state[] = 
{ 
  {"target", "Target", ISS_OFF},
};

/* Go to switch vector property */

static ISwitchVectorProperty indi_dome_goto_state_property = 
{
  MYDEV, 
  "goto", 
  "Go To ", 
  "", 
  IP_RW,
  ISR_NOFMANY, 
  0, 
  IPS_IDLE, 
  indi_dome_goto_state, 
  NARRAY(indi_dome_goto_state)
};

/* Rotate  switch  */

static ISwitch indi_dome_rotate_state[] = 
{ 
  {"ccw", "CCW", ISS_OFF},
  {"cw",  "CW", ISS_OFF}
};

/* Rotate state switch vector property */

static ISwitchVectorProperty indi_dome_rotate_state_property = 
{
  MYDEV, 
  "rotate", 
  "Rotate", 
  "", 
  IP_RW,
  ISR_NOFMANY, 
  0, 
  IPS_IDLE, 
  indi_dome_rotate_state, 
  NARRAY(indi_dome_rotate_state)
};


/* Upper shutter state switch */

static ISwitch indi_upper_shutter_state[] = 
{ 
  {"off", "Off", ISS_ON},
  {"open",  "Open",  ISS_OFF},
  {"close", "Close", ISS_OFF}
};

/* Upper shutter state switch vector property */

static ISwitchVectorProperty indi_upper_shutter_state_property = 
{
  MYDEV, 
  "upper_shutter", 
  "Upper Shutter", 
  "", 
  IP_RW,
  ISR_1OFMANY, 
  0, 
  IPS_IDLE, 
  indi_upper_shutter_state, 
  NARRAY(indi_upper_shutter_state)
};

/* Lower shutter state switch */

static ISwitch indi_lower_shutter_state[] = 
{ 
  {"off", "Off", ISS_ON},
  {"open",  "Open",  ISS_OFF},
  {"close", "Close", ISS_OFF}
};

/* Upper shutter state switch vector property */

static ISwitchVectorProperty indi_lower_shutter_state_property = 
{
  MYDEV, 
  "lower_shutter", 
  "Lower Shutter", 
  "", 
  IP_RW,
  ISR_1OFMANY, 
  0, 
  IPS_IDLE, 
  indi_lower_shutter_state, 
  NARRAY(indi_lower_shutter_state)
};


/* INDI functions */

/* Call once before any other IS function */
/* This turns on a timer */
/* When the timer expires it calls a camera routine */
/* The telescope routine will restart a new timer */

static void initdome()
{

  /* Initialization control */
  
  static int inited;
  
  if (inited)
  { 
    return;
  }
  
  /* Mounting and dome defaults */
  
  telmount = GEM;
  dec_offset = DECOFFSET;
  dec_height = DECHEIGHT;
  dec_length = DECLENGTH;
  dome_radius = DOMERADIUS;
  
  /* Longitude and latitude defaults */
  
  site_longitude = LONGITUDE;
  site_latitude = LATITUDE;
  
  /* For compatibility with previous versions of algorithms.c */
  
  SiteLongitude = site_longitude;
  SiteLatitude = site_latitude;
  
  /* Handle the configuration file */
    
  read_config();
  
  /* Set dome limits */
      
  if ( (upperlimit < 360.) || (lowerlimit > 0.) )
  {
    SetDomeLimits(lowerlimit, upperlimit);
    fprintf(stderr,"Warning: dome azimuth limits are %lf to %lf degrees.\n",
      lowerlimit, upperlimit);     
  }

  domeflag = FALSE;

  fprintf(stderr,"Connecting to the dome ... \n");
  
  /* Dome initialization and assignment of devices will go here */

  /* Tests that the dome is on line will go here */
  
  domeflag = TRUE; 
   
  if (domeflag != TRUE)
  {
    fprintf(stderr,"Dome controller is not available");
    exit(1);
  }
           
  /* Set flag so we do the initialization only once */
  
  inited = 1;

  /* Start the timer to poll the telescope */
  
  IEAddTimer (POLLMS, dome, NULL);
}

/* Send client the definitions of all properties we want it to know about */

void ISGetProperties (const char *dev)
{
   if (dev && strcmp (MYDEV, dev))
   {
     return;
   }
    
   initdome();
   IDDefNumber (&indi_dome_azimuth_property, NULL);
   IDDefNumber (&indi_dome_target_property, NULL);
   IDDefSwitch (&indi_dome_telaz_state_property, NULL);
   IDDefSwitch (&indi_dome_goto_state_property, NULL);
   IDDefSwitch (&indi_dome_rotate_state_property, NULL);
   IDDefSwitch (&indi_upper_shutter_state_property, NULL); 
   IDDefSwitch (&indi_lower_shutter_state_property, NULL);
}


/* Client is sending us a new blob */

void ISNewBLOB (const char *dev, const char *name, int sizes[],
  int blobsizes[], char *blobs[], char *formats[], char *names[], int n)
{
  /* This driver does not accept blobs */
}

/* Client is sending us new text */

void ISNewText (const char *dev, const char *name, 
  char *texts[], char *names[], int n)
{
  /* This driver does not accept text */
}


/* Client is sending us new switch settings */

/* In Elwood Downey's notation, the search arguments here are Sv not S */
/* In my notation the arguments are _property */

void  ISNewSwitch (const char *dev, const char *name, 
  ISState *states, char *names[], int n)
{
  int i;

  if (strcmp (dev, MYDEV))
  { 
      /* Ignore if this is not for us */
      return;
  }

  /* Handle request to monitor, or track telescope azimuth */
  
  if (!strcmp (name, indi_dome_telaz_state_property.name)) 
  {
    for (i = 0; i < n; i++) 
    {
      ISwitch *sp = IUFindSwitch (&indi_dome_telaz_state_property, names[i]);
      if (states[i] == ISS_ON) 
      {
        /* A request to poll the telescope was issued */
        
        for (i = 0; i < NARRAY(indi_dome_telaz_state); i++)
          indi_dome_telaz_state[i].s = ISS_OFF;
        sp->s = ISS_ON;
        if (!strcmp (sp->name, "off"))
        {          
          telpoll = TELPOLLOFF;
          indi_dome_telaz_state_property.s = IPS_IDLE;
          IDSetSwitch (&indi_dome_telaz_state_property, 
            "Dome not polling telescope"); 
        }
        else if (!strcmp (sp->name, "monitor"))
        {
          telpoll = TELPOLLON; 
          indi_dome_telaz_state_property.s = IPS_BUSY;
          IDSetSwitch (&indi_dome_telaz_state_property, 
            "Dome polling telescope");                    
        }
        else if (!strcmp (sp->name, "track"))
        {
          telpoll = TELPOLLTRACK; 
          indi_dome_telaz_state_property.s = IPS_ALERT;
          IDSetSwitch (&indi_dome_telaz_state_property, 
            "Dome tracking telescope");          
        }              
        else
        {      
          IDSetSwitch (&indi_dome_telaz_state_property, 
            "Unknown polling request");
        } 
        break;
      }
    }  
  }

  /* Handle a goto target request */
  /* Target coordinates must have already been loaded */
  /* When go to operation is complete the poll routine will change the flags */
  
  if (!strcmp (name, indi_dome_goto_state_property.name)) 
  {
    for (i = 0; i < n; i++) 
    {
      ISwitch *sp = IUFindSwitch (&indi_dome_goto_state_property, names[i]);
      if (states[i] == ISS_ON) 
      {
        /* A request to go to the target was issued */
        
        /* Ignore the request if dome tracking is enabled */
        
        if (telpoll == TELPOLLTRACK)
        {
          IDSetSwitch (&indi_dome_goto_state_property, 
            "Dome is tracking telescope"); 
          break;
        }  
        
        for (i = 0; i < NARRAY(indi_dome_goto_state); i++)
          indi_dome_goto_state[i].s = ISS_OFF;
        sp->s = ISS_ON;
        indi_dome_goto_state_property.s = IPS_BUSY;

        /* Set a flag reminding indi that a goto was started */
        
        domegoto = TRUE;
        
        /* Toggle the indi switch indiciating a goto is underway */
        
        IDSetSwitch (&indi_dome_goto_state_property, 
          "Dome rotating to target azimuth");
                
        /* Goto will be complete if domegoto == FALSE   */
        /* Indi will test the status in its normal poll cycle   */
        
        break;
      }
      if (states[i] == ISS_OFF) 
      {
        /* A go to target request was cancelled */
        
        for (i = 0; i < NARRAY(indi_dome_goto_state); i++)
          indi_dome_goto_state[i].s = ISS_ON;
        sp->s = ISS_OFF;
        indi_dome_goto_state_property.s = IPS_OK;
        
        /* Stop all rotation */
        
        ControlDome(DOMEOFF);
        domegoto = FALSE;
        
        IDSetSwitch (&indi_dome_goto_state_property, 
          "Dome azimuth rotation stopped");
            
        break;
      }      
    }  
  }

    
  /* Handle a rotation request by noting a change in the state property */
  /* Should also turn off a rotation when the property is deselected */

  if (!strcmp (name, indi_dome_rotate_state_property.name)) 
  {
    for (i = 0; i < n; i++) 
    {
      ISwitch *sp = IUFindSwitch (&indi_dome_rotate_state_property, names[i]);
      if (states[i] == ISS_ON) 
      {
        
        /* A request to rotate the dome was issued */        
        /* Unset any rotation state previously set */
        
        indi_dome_rotate_state[0].s = ISS_OFF;
        indi_dome_rotate_state[1].s = ISS_OFF;         
        
        /* Set the selected rotation state */
        
        sp->s = ISS_ON;

        /* Which direction have we selected? */
        
        if (!strcmp (sp->name, "cw"))
        {
          ControlDome(CWROTATION); 
        }
        if (!strcmp (sp->name, "ccw"))
        {
          ControlDome(CCWROTATION);
        }
           
        /* Indicate rotation is busy */
        
        indi_dome_rotate_state_property.s = IPS_BUSY;
        IDSetSwitch (&indi_dome_rotate_state_property, 
          "Starting rotation %s", sp->label);
        break;
      }
      if (states[i] == ISS_OFF) 
      {
        
        /* A rotation request was cancelled */
        
        sp->s = ISS_OFF;
        indi_dome_rotate_state_property.s = IPS_OK;
        ControlDome(STOPROTATION);           
        IDSetSwitch (&indi_dome_rotate_state_property, 
          "Stopping rotation %s", sp->label);

        break;
      }            
    }
  }


  /* Handle an upper shutter request by noting a change in the state property */
  /* Should also toggle shutter motion when another property is selected */

  if (!strcmp (name, indi_upper_shutter_state_property.name)) 
  {
    for (i = 0; i < n; i++) 
    {
      ISwitch *sp = IUFindSwitch (&indi_upper_shutter_state_property, names[i]);
      if (states[i] == ISS_ON) 
      {
        /* A request to change the upper shutter state was issued */
        
        for (i = 0; i < NARRAY(indi_upper_shutter_state); i++)
          indi_upper_shutter_state[i].s = ISS_OFF;
        sp->s = ISS_ON;
        indi_upper_shutter_state_property.s = IPS_OK;

        /* Which state have we selected? */
        
        if (!strcmp (sp->name, "off"))
        {          
          ControlDome(STOPUPPER);
        }
        else if (!strcmp (sp->name, "open"))
        {
          ControlDome(OPENUPPER); 
        }
        else if (!strcmp (sp->name, "close"))
        {
          ControlDome(CLOSEUPPER); 
        }              
        else
        {      
          IDSetSwitch (&indi_upper_shutter_state_property, 
            "Unknown upper shutter request");
          break;
        } 
          
        indi_upper_shutter_state_property.s = IPS_BUSY;
        
        IDSetSwitch (&indi_upper_shutter_state_property, 
          "Upper shutter %s", sp->label);
        break;
      }
    }
  }

  /* Handle a lower shutter request by noting a change in the state property */
  /* Should also toggle shutter motion when another property is selected */

  if (!strcmp (name, indi_lower_shutter_state_property.name)) 
  {
    for (i = 0; i < n; i++) 
    {
      ISwitch *sp = IUFindSwitch (&indi_lower_shutter_state_property, names[i]);
      if (states[i] == ISS_ON) 
      {
        /* A request to change the lower shutter state was issued */
        
        for (i = 0; i < NARRAY(indi_lower_shutter_state); i++)
          indi_lower_shutter_state[i].s = ISS_OFF;
        sp->s = ISS_ON;
        indi_lower_shutter_state_property.s = IPS_OK;

        /* Which state have we selected? */
        
        if (!strcmp (sp->name, "off"))
        {          
          ControlDome(STOPLOWER);
        }
        else if (!strcmp (sp->name, "open"))
        {
          ControlDome(OPENLOWER); 
        }
        else if (!strcmp (sp->name, "close"))
        {
          ControlDome(CLOSELOWER); 
        }              
        else
        {      
          IDSetSwitch (&indi_lower_shutter_state_property, 
            "Unknown lower shutter request");
          break;
        } 
          
        indi_lower_shutter_state_property.s = IPS_BUSY;
        
        IDSetSwitch (&indi_lower_shutter_state_property, 
          "Upper shutter %s", sp->label);
        break;
      }
    }
  }
}

/* Client is sending new numbers */
/* Search number properties for a match and act accordingly */

/* In Elwood Downey's notation, the search arguments here are Nv not N */
/* In my notation the arguments are _property */

void ISNewNumber (const char *dev, const char *name,
  double *doubles, char *names[], int n)
{
  
  if (strcmp (dev, MYDEV))
  { 
      /* Ignore if this is not for us */
      return;
  }

  if (!strcmp (name, indi_dome_target_property.name) && n > 0) 
  {    
    
    /* A new target has been sent */
    
    indi_dome_target[0].value = doubles[0];
    dometargetaz = indi_dome_target[0].value;
   
    IDSetNumber (&indi_dome_target_property, 
      "Target Azimuth %.0f", dometargetaz);    
  }

}

/* Dome function called during the polling cycle */
/* Restart the timer to call this function again */

static void dome (void *p)
{
  static int tcount;
  int nloops;
  double telra, teldec;
    
  /* Local very high priority operations would go here */
  
  if (domegoto == TRUE)
  {
    domegoto = DomeToAzimuth(dometargetaz);
    if ( domegoto == FALSE )
    {
       /* Target goto to is complete */
       
       indi_dome_goto_state[0].s = ISS_OFF; 
       indi_dome_goto_state_property.s = IPS_OK;      
       IDSetSwitch (&indi_dome_goto_state_property, "Goto complete");        
       fprintf(stderr,"domegoto %d dometargetaz %lf\n", 
         domegoto, dometargetaz);
    }    
  }  
  
  /* Otherwise poll dome every 16 seconds for a 250 ms INDI polling cycle */
  
  if ((tcount < 0) || (tcount > 64))
  {
    tcount = 1;  
   
    /* Get the latest dome azimuth and state */
    
    GetDome(&domeaz,&domestatus);
    GetShutter(&upper_shutter_status,&lower_shutter_status); 
    
    /* Local mid-priority operations would go here */
        
    if (telpoll != TELPOLLOFF)
    {
      dometargetaz = 180.;
      nloops = 4;
      GetTel(&telra, &teldec);
      SolveDomeAzimuth(telra, teldec, &dometargetaz, nloops);
      
      indi_dome_target[0].value = dometargetaz;
      IDSetNumber (&indi_dome_target_property, 
        "Target Azimuth %.0f", dometargetaz); 
      if (telpoll == TELPOLLTRACK)
      {
        if ( (dometargetaz >= lowerlimit) && (dometargetaz <= upperlimit) )
        {
          /* Rotate dome to this azimuth within a 5 degree tolerance */
          domegoto = DomeToAzimuth(dometargetaz);
        }  
      }                      
    }
    
    /* Update the server */

    indi_dome_azimuth[0].value = domeaz;
    IDSetNumber (&indi_dome_azimuth_property, NULL);    

  }
  tcount++;
  
  /* Start the timer again */
  IEAddTimer (POLLMS, dome, NULL);
}


/* Read and parse the initial configuration file */

void read_config(void)
{
  char configstr[121];
  char *configptr = configstr;
  char *configfile;  
  FILE *fp_config; 
  
  configfile = (char *) malloc (MAXPATHLEN);
  strcpy(configfile,CONFIGFILE);  
  fp_config = fopen(configfile, "r");
  
  if ( fp_config == NULL )
  {
    fprintf(stderr,"Caution: required dome configuration not found.\n");
    return;
  }
  
  while ( configstr == fgets(configstr,80,fp_config) )
  {
    
    configptr = strstr(configstr,"dome.longitude");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &site_longitude);
        SiteLongitude = site_longitude;
      }  
    }
    configptr = strstr(configstr,"dome.latitude");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &site_latitude);
        SiteLatitude = site_latitude;
      }  
    }
    configptr = strstr(configstr,"dome.mount");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%d", &telmount);
      }  
    }
    configptr = strstr(configstr,"dome.decoffset");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_offset);
      }  
    }
    configptr = strstr(configstr,"dome.decheight");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_height);
      }  
    }
    configptr = strstr(configstr,"dome.declength");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_length);
      }  
    }
    configptr = strstr(configstr,"dome.radius");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dome_radius);
      }  
    }
    configptr = strstr(configstr,"dome.upperlimit");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &upperlimit);
      }  
    }
    configptr = strstr(configstr,"dome.lowerlimit");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &lowerlimit);
      }  
    }
  }
     
  fclose(fp_config);
}     
      

