/* Dome Control Protocol for the University of Louisville's RFID Tag System  */
/*                                                                           */
/* Copyright 2010 Jeff Hay and John Kielkopf                                 */
/*                                                                           */
/* Distributed under the terms of the General Public License (see LICENSE)   */
/*                                                                           */
/* John Kielkopf (kielkopf@louisville.edu)                                   */
/* June 16, 2010 -- Version 1.3.9                                            */
/*                                                                           */
/*                                                                           */
/* History:                                                                  */
/*                                                                           */
/*                                                                           */
/* August 29, 2008                                                           */
/*   Version 1.0                                                             */
/*                                                                           */
/*   Control of dome rotation                                                */
/*   National Instruments PCI-6518 parallel IO for rotation switch control   */
/*   Phidget RFID tags and USB tag reader for encoding absolute rotation     */
/*                                                                           */
/*   Modeled Jeff Hay's domecontrol.c                                        */
/*   Successful operation of dome rotation through INDI                      */
/*                                                                           */
/* July 30, 2009                                                             */
/*   Version 1.1                                                             */
/*                                                                           */
/*   Angle read from the system file /tmp/domeazimuth                        */
/*   Goto implemented                                                        */
/*                                                                           */
/* August 8, 2009                                                            */
/*   Version 1.2                                                             */
/*                                                                           */
/*   Dome auto track telescope RA and Dec added                              */
/*                                                                           */
/* September 26, 2009                                                        */
/*   Version 1.3                                                             */
/*                                                                           */
/*   Modified search algorithm                                               */
/*                                                                           */
/* November 24, 2009                                                         */
/*   Version 1.3.6                                                           */
/*                                                                           */
/*   Corrected coding errors in solve azimuth                                */
/*                                                                           */
/* May 31, 2010                                                              */
/*   Version 1.3.7                                                           */
/*                                                                           */
/*   Corrected southern hemisphere layout error                              */
/*                                                                           */
/* June 6, 2010                                                              */
/*   Version 1.3.9                                                           */
/*   Modified southern hemisphere calculation                                */
/*   Changed tel and dome files to /usr/local/observatory/status/            */
                                                                             
                                                                             
#include <stdio.h>                                                           
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include <NIDAQmx.h>
#define DAQmxErrChk(functionCall) if( DAQmxFailed(error=(functionCall)) ) goto Error; else  

#include "protocol.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/* Flags save current state */

static int domestatus = 0;
static int domegoto = FALSE;
static int lowerstatus = 0;
static int upperstatus = 0;
static double domeazimuth = 0.;
static double lowerlimit = 0.;
static double upperlimit = 360.;
static double domedegpersec = DOMEDEGPERSEC;
static double rotatesecs; 
static struct timeval tv, tv0; 
 
/* Prototypes for dome control */

void ControlDome(int domecmd);
int  DomeToAzimuth(double dometargetaz); 
void SolveDomeAzimuth(double ra, double dec, double *az, int nloops);
void GetDome(double *domeaz, int *domestatus);
void GetShutter( int *upper, int *lower);
void GetTel(double *ra, double *dec);
void GetShutterLimits(void);
void SetDomeLimits(double lower, double upper);

/* Prototypes for dome azimuth synchronization */

extern double Map24(double hour);
extern double Map12(double hour);
extern double Map360(double angle);
extern double Map180(double angle);
extern double frac(double x);
extern double LSTNow(void);
extern void EquatorialToHorizontal(double ha, double dec, double *az, double *alt);
extern void HorizontalToEquatorial(double az, double alt, double *ha, double *dec);

/* Global variables for dome azimuth synchronization */
/* Set these in a calling program through a configuration file */

extern double site_longitude; 
extern double SiteLongitude; 
extern double site_latitude;
extern double SiteLatitude;

extern double dec_offset;  /* Center of dome to dec axis */   
extern double dec_height;  /* Height of dec axis at polar axis */
extern double dec_length;  /* Length from polar to optical axes */
extern double dome_radius; /* Radius of dome */
extern int telmount;

/* Files */

FILE *fp_azimuth;
static char *azimuthfile;
FILE *fp_telcoords;
static char *telcoordsfile;


/* Request a dome operation */

void ControlDome(int domecmd)
{
  int32 error=0;
  TaskHandle taskHandle=0;
  uInt8 data[8];
  static unsigned int lastdata[8];
  char errBuff[2048]={'\0'};

  if (domecmd == DOMEOFF)
  {
    lastdata[0]= 0;
    lastdata[1]= 0;
    lastdata[2]= 0;
    lastdata[3]= 0;
    lastdata[4]= 0;
    lastdata[5]= 0;
    lastdata[6]= 0;
    lastdata[7]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];   
    domegoto = FALSE;
    domestatus =  ( domestatus & ~(GOTOINPROGRESS) );    
    domestatus =  ( domestatus  & ~(ROTATINGCW)     );
    domestatus =  ( domestatus  & ~(ROTATINGCCW)    );
    upperstatus = ( upperstatus  & ~(OPENING)        );
    upperstatus = ( upperstatus & ~(CLOSING)        );
    lowerstatus = ( lowerstatus & ~(OPENING)        );
    lowerstatus = ( lowerstatus & ~(CLOSING)        );
  }
  else if (domecmd == CWROTATION)
  {
    /* Rotate dome clockwise decreasing its azimuth */
    domestatus = ( domestatus | ROTATINGCW );
    lastdata[0]= 1;
    lastdata[1]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];      
    
  }
  else if (domecmd == CCWROTATION)
  {
    /* Rotate dome counterclockwise increasing its azimuth */
    domestatus =   ( domestatus | ROTATINGCCW ); 
    lastdata[0]= 0;
    lastdata[1]= 1;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];  
  }
  else if (domecmd == STOPROTATION)
  {
    /* Stop dome rotation */
    domestatus =  ( domestatus  & ~(ROTATINGCW) );
    domestatus =  ( domestatus  & ~(ROTATINGCCW) ); 
    lastdata[0]= 0;
    lastdata[1]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];    
  } 
  else if (domecmd == OPENUPPER)
  {
    /* Open upper shutter */
    upperstatus =  ( upperstatus & OPENING );
    upperstatus =  ( upperstatus & NOTCLOSED );
    lastdata[2]= 1;
    lastdata[3]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];     
  } 
  else if (domecmd == CLOSEUPPER)
  {
    /* Close upper shutter */
    upperstatus = ( upperstatus & CLOSING );
    lastdata[2]= 0;
    lastdata[3]= 1;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];     
  }
  else if (domecmd == STOPUPPER)
  {
    /* Stop upper shutter */
    upperstatus = ( upperstatus & ~(OPENING) );
    upperstatus = ( upperstatus & ~(CLOSING) );
    lastdata[2]= 0;
    lastdata[3]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];    
  }  
  else if (domecmd == OPENLOWER)
  {
    /* Open lower shutter */
    lowerstatus = ( lowerstatus & OPENING );
    lowerstatus = ( lowerstatus & NOTCLOSED );
    lastdata[4]= 1;
    lastdata[5]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];    
  } 
  else if (domecmd == CLOSELOWER)
  {
    /* Close lower shutter */
    lowerstatus = ( lowerstatus & CLOSING );
    lastdata[4]= 0;
    lastdata[5]= 1;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];     
  }  
  else if (domecmd == STOPLOWER)
  {
    /* Stop lower shutter */
    lowerstatus = ( lowerstatus & ~(OPENING) );
    lowerstatus = ( lowerstatus & ~(CLOSING) );
    lastdata[4]= 0;
    lastdata[5]= 0;
    data[0]= lastdata[0];
    data[1]= lastdata[1];
    data[2]= lastdata[2];
    data[3]= lastdata[3];
    data[4]= lastdata[4];
    data[5]= lastdata[5];
    data[6]= lastdata[6];
    data[7]= lastdata[7];     
  }    
  else
  {
    fprintf(stderr,"Unknown dome command\n");
    return;
  } 
  
  /* DAQmx Configure Code */
  /* This is where the device, line and port are set */
  /* It is probably Dev1 if no other National Instruments card is present */
  /* Port 2 is determined by the wiring and is correct for the Moore dome */
    
  DAQmxErrChk (DAQmxCreateTask("",&taskHandle));
  DAQmxErrChk (DAQmxCreateDOChan(taskHandle,
    "Dev1/port2/line0:7","",DAQmx_Val_ChanForAllLines));

  /* DAQmx Start Code */
  
  DAQmxErrChk (DAQmxStartTask(taskHandle));
  
  /* DAQmx Write Code */

  DAQmxErrChk (DAQmxWriteDigitalLines(taskHandle,1,1,10.0,
    DAQmx_Val_GroupByChannel,data,NULL,NULL));

  /* DAQmx Error Code */
  
  Error:
    if( DAQmxFailed(error) )
      DAQmxGetExtendedErrorInfo(errBuff,2048);

  if( taskHandle!=0 ) 
  {

    /* DAQmx Stop Code */

    DAQmxStopTask(taskHandle);
    DAQmxClearTask(taskHandle);
  }

  if( DAQmxFailed(error) )
  {
    fprintf(stderr,"DAQmx Error: %s\n",errBuff);
  }

  return;
}

/* Rotate the dome toward the target azimuth */
/* Same as XEphem - CCW looking up increases azimuth */
/* Returns domegoto flag */
/* Make repeated calls with same target to complete goto */
/* Cancel with a call using an out-of-range azimuth */
/*   or send dome a DOMEOFF request */

int DomeToAzimuth(double targetazimuth)
{
  double nowazimuth;
  double delta;
  int nowstate;
  int rotatedir = CWROTATION;
  double dt;

  /* Assure target is between 0. to +360. */
   
  if ( (targetazimuth >= 360.) || (targetazimuth <= 0.) )
  {
    domegoto = FALSE;
    domestatus = ( domestatus & ~(GOTOINPROGRESS) );
    ControlDome(STOPROTATION);
    return (domegoto);
  }
  
  /* Assure target is between lower and upper limits */
   
  if ( (targetazimuth >= upperlimit) || (targetazimuth <= lowerlimit) )
  {
    domegoto = FALSE;
    domestatus = ( domestatus & ~(GOTOINPROGRESS) );
    fprintf(stderr,"Request for dome azimuth outside allowed limits.\n");
    return (domegoto);
  } 
       
  if (domegoto == TRUE)
  {

    /* We already have a goto in progress */
    /* Check if it has taken longer that we expected */
        
    gettimeofday(&tv, NULL);
    dt = tv.tv_sec - tv0.tv_sec;
    
    if (dt > rotatesecs)
    {
      domegoto = FALSE;
      domestatus = ( domestatus & ~(GOTOINPROGRESS) );
      ControlDome(STOPROTATION);
      return (domegoto);
    }

    /* Is it close to the target? */
    
    GetDome(&nowazimuth, &nowstate);
   
    /* Safety check - halt slew if the dome is within the exclusion zone */
    /* Assumes nowazimuth falls between 0 and 360 */
    
    if ( (nowazimuth > upperlimit) || ( nowazimuth < lowerlimit) )
    {
      domegoto = FALSE;
      domestatus = ( domestatus & ~(GOTOINPROGRESS) );
      ControlDome(STOPROTATION); 
      fprintf(stderr, "Current azimuth outside allowed range.\n");
      return (domegoto);   
    }
    
    /* How much more to rotate? */
        
    delta = targetazimuth - nowazimuth;
    
    /* Use +- 180 */
    
    if ( delta >= 180. )
    {  
      delta = delta - 360.;
    }
    else if ( delta <= -180. )
    {
      delta = delta + 360.;
    }
     
    /* Is it within 5 degree tolerance? */
    
    if ( (delta > -6.) && (delta < 6.) )
    {
      domegoto = FALSE;
      domestatus = ( domestatus & ~(GOTOINPROGRESS) );
      ControlDome(STOPROTATION);
    }  
    return (domegoto);
  }
  
  /* A goto was not in progress so we start one */
  /* Reset goto flag true */
  
  domegoto = TRUE;
  domestatus = ( domestatus & GOTOINPROGRESS ) ;
    
  /* Set the starting time */
  
  gettimeofday(&tv0, NULL);
  
  /* Check on current azimimth */
  
  GetDome(&nowazimuth, &nowstate);
  
  /* How much to rotate */
      
  delta = targetazimuth - nowazimuth;
  
  /* Use +- 180 */
  
  if ( delta >= 180. )
  {  
    delta = delta - 360.;
  }
  else if ( delta <= -180. )
  {
    delta = delta + 360.;
  }
   
  /* Within 5 degree tolerance? */
  
  if ( (delta > -6.) && (delta < 6.) )
  {
    domegoto = FALSE;
    domestatus = ( domestatus & ~(GOTOINPROGRESS) );
    return (domegoto);
  }   

  /* Set direction and estimate rotation time */
  
  if ( delta > 0. )
  {
    rotatedir = CCWROTATION;
    rotatesecs = delta/domedegpersec;
  }
  else
  {
    rotatedir = CWROTATION;    
    rotatesecs = -delta/domedegpersec;
  }
  rotatesecs = 1. + rotatesecs;
     
  /* Start dome rotation */
  
  ControlDome(rotatedir);
  
  /* Repeat calls until tolerance met or timeout */

  return (domegoto);
}

/* Solve for the azimuth of the dome given the telescope RA and Dec */

void SolveDomeAzimuth(double ra, double dec, double *az, int nloops)
{
  double ha, ha0, telaz, telalt;
  double R;             /* Dome radius */
  double H, L, P;       /* Shorthand for mount parameters */   
  double x0, y0, z0;    /* Reference point on the optical axis */
  double x, y, z;       /* Hypothetical intersection with the dome surface */
  double rp, r;         /* Distance to dome from OTA reference point */ 
  double phi;           /* Rotation of the polar axis in degrees */
  double theta;         /* Angle of polar axis to the horizontal */
  double telaz2;        /* Southern hemisphere azimuth from south in radians */
  double telalt2;       /* Southern hemisphere altitude in radians */
  
  double zeta;          /* Dome azimuth */
  int n;                /* Iteration loop counter */
  
  /* Initialize x,y,z,r */
  
  x = 0.;
  y = 0.;
  z = 0.;
  x0 = 0.;
  y0 = 0.;
  z0 = 0.;
  r = 0;
  
  /* Radius of dome */

  R = dome_radius;
  
  /* Offset to dec axis */
  
  P = dec_offset;
  
  /* Height of intersection of declination and polar axes */
  
  H = dec_height;
  
  /* Distance from polar to optical axes */
  
  L = dec_length;
  
  /* Compute the current HA given the RA and the sidereal time */
  /* This is valid in either hemisphere */
  
  ha = LSTNow();
  ha = ha - ra;
  ha = Map24(ha);
  ha0 = Map12(ha);
  
  /* Find the altitude and azimuth of the current pointing            */
  /* Function requires ha in 0 to 24 basis                            */
  /* This should be valid in either hemisphere                        */
  
  EquatorialToHorizontal(ha, dec, &telaz, &telalt);
    
  /* Find the reference point on the optical axis in dome coordinates */
  /* z: vertical                                                      */
  /* y: north - south                                                 */
  /* x: east - west                                                   */
  /* theta: altitude of the polar axis from the horizontal plane      */
  /* phi: rotation of dec axis about polar axis                       */
    
  /* Meaning of x and y compared to sky depends on the hemisphere!    */
  /* z: always + up                                                   */
  /* y: always + toward N for +lat or S for -lat                      */  
  /* x: maintains right-handed coordinate system with z and y         */
  /*  thus x is + to E for +lat and to W for -lat                     */
  /* theta: always positive from the horizontal plane to the pole     */
  /* phi: rotation about polar axis                                   */
  /*  and is +90 for a horizontal axis with OTA toward +x             */
  /*  and is   0 for OTA over mount with dec axis down                */
  /*  and is -90 for a horizontal axis with OTA toward -x             */
  
  if (telmount != GEM)
  {
    x0 = 0.;
    y0 = P;
    z0 = H;
  }
  else 
  {
            
    /* Assign phi */
    
    if (site_latitude >= 0.)
    {
      if (ha0 > 0.)
      {
        /* Looking west with OTA east of pier */
        /* HA is between 0 and +12 hours */

        phi = (6. - ha0)*PI/12.;
        
      }
      else
      {  
        /* Looking east with OTA west of pier */
        /* HA is between -12 and 0 hours */
        
        phi = -(6. + ha0)*PI/12.;
        
      }
    }
    else
    {
      if (ha0 > 0.)
      {
        /* Looking west with OTA east of pier */
        /* HA is between 0 and +12 hours */
        
        phi = -(6. - ha0)*PI/12.;
        
      }
      else
      {  
        /* Looking east with OTA west of pier */
        /* HA is between -12 and 0 hours */
        
        phi = (6. + ha0)*PI/12.;
        
      }    
    }
    
    /* Assign theta */
        
    if (site_latitude >= 0.)
    {
      theta = site_latitude*PI/180.;
    }
    else
    {
      theta = -1.*site_latitude*PI/180.;
    }
 
    /* Find the dome coordinates of the OTA reference point */
    /* This works in either hemisphere                      */
         
    x0 = L*sin(phi);
    y0 = -L*cos(phi)*sin(theta) + P; 
    z0 = L*cos(phi)*cos(theta) + H;
  }
    
  /* (x,y,z) is on the optical axis */
  /* Iterate to make this point also lie on the dome surface */
  
  /* Begin iteration assuming the zero point is at the center of the dome */
  /* Telescope azimuth is measured from the direction to the pole */
  
  rp = R;

  if (site_latitude >= 0.)
  {
    telaz2 = telaz * PI / 180.;
    telalt2 = telalt * PI / 180.;
  }
  else
  {
    telaz2 = telaz - 180.;
    telaz2 = Map360(telaz2); 
    telaz2 = telaz2 * PI / 180.;
    telalt2 = telalt * PI / 180.;       
  }
  
  for ( n=0; n<nloops; n++)  
  {
    x = x0 + rp * sin(telalt2) * sin(telaz2);
    y = y0 + rp * sin(telalt2) * cos(telaz2);
    z = z0 + rp * cos(telalt2);
    r = sqrt(x*x + y*y + z*z);
    
    /* Repeat the calculation correcting the assumed zero point each time */
    /* This converges quickly  so just do a few loops and assume it is ok */
    
    rp = rp - 0.5*(r - R);
  }
    
  /* Use (x,y,0) from the interation to find the azimuth of the dome */
  /* Azimuth is N (0), E (90), S (180), W (270) in both hemispheres */
  /* However x and y are different in the hemispheres so we fix that here */
  
  zeta  = atan2(x,y);
  if ( (zeta > - 2.*PI) && (zeta < 2.*PI))
  {
    if ( site_latitude > 0. )
    {
      zeta = (180./PI)*zeta;
      zeta = Map360(zeta);
    }
    else 
    {
      zeta = (180./PI)*zeta;
      zeta = zeta + 180.;
      zeta = Map360(zeta);   
    }  
  }
  else
  {
    zeta = telaz;
  }
  
  *az = zeta;
  
}


/* Return the dome azimuth and state */
/* Note azimuth file is an integer but internal domeazimuth is double */

void GetDome(double *az, int *status)
{
  int current_azimuth;
  int nread;
  
  azimuthfile = (char*) malloc (MAXPATHLEN);
  strcpy(azimuthfile,AZIMUTHFILE);
  fp_azimuth = fopen(azimuthfile, "r");
  if (fp_azimuth == NULL)
  {
    return;
  }
  
  nread = fscanf(fp_azimuth, "%d", &current_azimuth);
  fclose(fp_azimuth);
  
  if (nread !=0)
  {
    domeazimuth = current_azimuth;
    *az = domeazimuth;
    *status = domestatus;
    return;
  }

  /* Returns last known state from prior operations if read unsuccessful */
  
  *az = domeazimuth;
  *status = domestatus;

  return;
}

/* Return the shutter status */

void GetShutter(int *upper, int *lower)
{
  
  GetShutterLimits();
  
  /* Returns last known state from prior operations */
  
  *upper = upperstatus;
  *lower = lowerstatus;

  return;
}

/* Test shutter limits and update status */

void GetShutterLimits(void)
{  
  
  /* Holding place for sensor check of dome shutter */
  
  return;
}

/* Accept limits on dome rotation */
/* Might be used if dome shutter is connected by a drop cable */

void SetDomeLimits(double lower, double upper)
{  
  if ( upper < 0. )
  {
    fprintf (stderr,"Upper azimuth limit cannot be negative.\n");
    return;
  } 
  if ( lower < 0. )
  {
    fprintf (stderr,"Lower azimuth limit cannot be negative.\n");
    return;
  }
  if ( upper >360. )
  {
    fprintf (stderr,"Upper azimuth limit cannot be larger than 360 degrees.\n");
    return;
  } 
  if ( lower > 360. )
  {
    fprintf (stderr,"Lower azimuth limit cannot be larger than 360 degrees.\n");
    return;
  }     
  if (upper > lower)
  {
    lowerlimit = lower;
    upperlimit = upper;
  }
  else
  {
    fprintf(stderr,
      "Upper azimuth limit must be larger than lower lower limit.\n");
  }
  return;
}


/* Get the telescope coordinates with ra in hours and dec in degrees */
/* Returns TRUE if coordinates are valid, FALSE if not */

void GetTel(double *ra, double *dec)
{
  int nread;
  double current_ra, current_dec;
  
  telcoordsfile = (char*) malloc (MAXPATHLEN);
  strcpy(telcoordsfile,TELCOORDSFILE);
  fp_telcoords = fopen(telcoordsfile, "r");
  if (fp_telcoords == NULL)
  {
    return;
  }
  
  nread = fscanf(fp_telcoords, "%lf %lf", &current_ra, &current_dec);
  fclose(fp_telcoords);
  
  if (nread !=0)
  {
    *ra = current_ra;
    *dec = current_dec;
  }


  return;
}
