
/* ----------------------------------------------------------------------- */
/* -                       Astronomical Dome Control                     - */
/* -                         Motif User Interface                        - */
/* -                         Standalone Operation                        - */
/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Copyright (c) 2009 John Kielkopf                                        */
/* kielkopf@louisville.edu                                                 */
/*                                                                         */
/* This file is part of XmDome.                                            */
/*                                                                         */
/* Distributed under the terms of the General Public License (see LICENSE) */
/*                                                                         */
/* Date: November 27, 2009                                                 */
/* Version: 1.3.7                                                          */
  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h> 
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/Separator.h>
#include <Xm/SelectioB.h>

#include "xmdome1.h"
#include "protocol.h"

/* Dependencies */

/* protocol.c   */
/* algorithms.c */

/* Dome parameters */

double site_longitude;     /* Preferred naming convention */
double site_latitude;      /* Preferred naming convention */
double SiteLongitude;      /* Needed by algorithms */
double SiteLatitude;       /* Needed by algorithms */
double dec_offset;         /* Center of dome to dec axis */   
double dec_height;         /* Height of dec axis at polar axis */
double dec_length;         /* Length from polar to optical axes */
double dome_radius;        /* Radius of dome */
int telmount = GEM;        /* One of GEM (2) , EQFORK (1) , ALTAZ (0) */
double upperlimit = 360.;  /* Maz azimuth for dome rotation */
double lowerlimit = 0.;    /* Min azimuth for dome rotation */
  
/* User interface */

XtAppContext context;
XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;

Widget menu_bar;

Widget file_menu;
Widget file_exit_item;

Widget set_menu;
Widget set_target_item;

Widget toplevel;

Widget target_dialog;

Widget button[3][3];            
Widget command[12];
Widget toggle[2];
Widget label_info[6];
Widget show_info[6];
Widget form;
Widget sep_btns;
Widget sep_status;
Widget radio_box;

Widget make_menu();              /* creates a menu on the menu bar */
Widget make_menu_item();         /* adds an item into the menu */


/* labels for the buttons top to bottom then left to right */

static char btn_name[][8] = {
  "NoOp",
  "CW",
  "NoOp",
  "NoOp", 
  "Stop",
  "NoOp",
  "NoOp",
  "CCW",
  "NoOp"
};


static char cmd_name[][16] = {
  "Set Telescope",
  "Start Slew",
  "Stop Slew",  
  "Set North",
  "Set South",
  "Set Flat",  
  "Open Upper",
  "Close Upper",
  "Stop Upper",  
  "Open Lower",  
  "Close Lower",
  "Stop Lower"
}; 
  

static char trk_name[][8] = {
  "Manual",
  "Auto"
 };

static char infotext[][8] = {
  "Azimuth",
  "Target",
  "CW",
  "CCW",
  "Upper",
  "Lower"
};

static char infodata[][24] = {
  "-",
  "-",
  "-",
  "-",
  "-",
  "-"
};

void create_menus(Widget menu_bar);  /* Creates all menus                 */
void menuCB();                       /* Callback for all pull-down menus  */
void target_dialogCB();              /* Callback for target dialog        */

void buttonCB();           /* Callback function for the paddle buttons    */
void toggleCB();           /* Callback function for the toggles           */
void commandCB();          /* Callback function for command buttons       */

/* Manage the dome interface */

void initdome(void);
void dome(void);
void read_config(void);

/* Parameter reporting */

void update_azimuth_display();       /* Display dome azimuth        */
void update_target_display();        /* Display target azimuth      */
void update_cw_display();            /* Display cw state            */
void update_ccw_display();           /* Display ccd state           */
void update_upper_display();         /* Display upper shutter state */
void update_lower_display();         /* Display lower shutter state */
void update_all_displays();          /* Display all buffers         */

/* Buffers */

char buf[256];                   /* general purpose character buffer */

/* Internal control flags */

int domestatus = 0;
int domecmd = DOMEOFF;
int domegoto = FALSE;
int upper_shutter_status = 0;
int lower_shutter_status = 0;
int panelautotrack = FALSE;

/* Dome azimuth */

double domeaz = 0.;
double dometargetaz = 0.;

/* User interface polling */

unsigned long poll_interval = POLLMS;      
void poll_interval_handler(XtPointer client_data_ptr, XtIntervalId *client_id);            
XtIntervalId poll_interval_id;
XtPointer poll_interval_data_ptr;


/* Functions from the dome protocol */

extern void ControlDome(int domecmd);
extern int  DomeToAzimuth(double dometargetaz); 
extern void SolveDomeAzimuth(double ra, double dec, double *az, int nloops);
extern void GetDome(double *domeaz, int *state);
extern void GetShutter(int *upperstatus, int *lowerstatus);
extern void GetTel(double *ra, double *dec);
extern void SetDomeLimits(double lower, double upper);

/* GUI main starts here */

int main(int argc, char *argv[])
{
  Arg al[10];
  int ac;
  int x,y,btn_number,cmd_number;
  int info_number;
  
  EventMask mask;

  /* Contact the dome */
  
  initdome();
  
  /* Create the toplevel shell */

  toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
      NULL,NULL,0);
  
  /* Set the default size of the window */ 

  ac=0;
  XtSetArg(al[ac],XmNwidth,400); ac++;
  XtSetArg(al[ac],XmNheight,700); ac++;
  XtSetValues(toplevel,al,ac);

  /* Create a form widget */

  ac=0;
  form=XmCreateForm(toplevel,"form",al,ac);
  XtManageChild(form);

  /* Create the menu bar and attach it to the form */
  
  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
  menu_bar=XmCreateMenuBar(form,"menu_bar",al,ac);
  XtManageChild(menu_bar);

  /* Locations of buttons are in percent of window dimension */

  /* Create a radio box to hold the toggles and attach it to form */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmHORIZONTAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,5); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,35); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,90); ac++; 
  radio_box=XmCreateRadioBox(form,"radio_box",al,ac);
  XtManageChild(radio_box);

  /* Create toggles */
  
  for (x=0; x<2; x++)
  {
    ac=0;
    XtSetArg(al[ac],XmNlabelString,
        XmStringCreate(trk_name[x],char_set)); ac++;
    toggle[x]=XmCreateToggleButton(radio_box,"toggle",al,ac);
    XtManageChild(toggle[x]);
    XtAddCallback (toggle[x], 
      XmNvalueChangedCallback, toggleCB, (int *) x);
  }     


  /* Set the togglebuttons to initial values */

  XmToggleButtonSetState (toggle[0], True, False);

  /* Set up the handpaddle buttons, add event handlers, attach to form */
  /* Nine buttons are numbered from 0 to 8 */
  
  for (x=0; x<3; x++)
  {
    for (y=0; y<3; y++)
    {
      btn_number=3*x+y;
      if ( (btn_number==1) | (btn_number==4) | (btn_number==7) ) 
      {
        ac=0;
        XtSetArg(al[ac],XmNlabelString,
            XmStringCreate(btn_name[btn_number],char_set)); ac++;
        XtSetArg(al[ac],XmNleftAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNleftPosition,27+x*16); ac++;
        XtSetArg(al[ac],XmNrightAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNrightPosition,41+x*16); ac++;
        XtSetArg(al[ac],XmNtopAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNtopPosition,4+y*9); ac++;
        XtSetArg(al[ac],XmNbottomAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNbottomPosition, 12+y*9); ac++;
        button[x][y]=XmCreatePushButton(form,"label",al,ac);
        mask = ButtonReleaseMask | ButtonPressMask;
        XtAddEventHandler(button[x][y],mask,False,
            buttonCB,(int *) btn_number);              
        XtManageChild(button[x][y]);
      }
    }
  }



  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,28); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;  
  sep_btns=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_btns);   


  /* Set up the information label widgets and attach them to the form */

  for  (x=0; x<2; x++)
  {
    for (y=0; y<3; y++)
    {
      info_number=2*y+x;
      ac=0;
      XtSetArg(al[ac],XmNlabelString,
        XmStringCreate(infotext[info_number],char_set)); ac++;
      XtSetArg(al[ac],XmNleftAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNleftPosition,5+x*50); ac++;
      XtSetArg(al[ac],XmNrightAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNrightPosition,45+x*50); ac++;
      XtSetArg(al[ac],XmNtopAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNtopPosition,34+y*6); ac++;
      XtSetArg(al[ac],XmNbottomAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNbottomPosition,36+y*6); ac++;
      label_info[info_number]=XmCreateLabel(form,"label",al,ac);
      XtManageChild(label_info[info_number]); 
    }
  }


  /* Set up the information data widgets and attach them to the form */

  for  (x=0; x<2; x++)
  {
    for (y=0; y<3; y++)
    {
      info_number=2*y+x;
      ac=0;
      XtSetArg(al[ac],XmNlabelString,
        XmStringCreate(infodata[info_number],char_set)); ac++;
      XtSetArg(al[ac],XmNleftAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNleftPosition,5+x*50); ac++;
      XtSetArg(al[ac],XmNrightAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNrightPosition,45+x*50); ac++;
      XtSetArg(al[ac],XmNtopAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNtopPosition,36+y*6); ac++;
      XtSetArg(al[ac],XmNbottomAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNbottomPosition,38+y*6); ac++;
      show_info[info_number]=XmCreateLabel(form,"label",al,ac);
      XtManageChild(show_info[info_number]); 
    }
  }


  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,54); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_status=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_status);   

  /* Set up the command buttons and attach them to the form */

  for (x=0; x<4; x++)   
  {
    for (y=0; y<3; y++)
    {
      cmd_number=3*x+y;
      ac=0;
      XtSetArg(al[ac],XmNlabelString,
        XmStringCreate(cmd_name[cmd_number],char_set)); ac++;
      XtSetArg(al[ac],XmNleftAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNleftPosition,3+x*24); ac++;
      XtSetArg(al[ac],XmNrightAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNrightPosition,25+x*24); ac++;
      XtSetArg(al[ac],XmNtopAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNtopPosition,64+y*9); ac++;
      XtSetArg(al[ac],XmNbottomAttachment,
        XmATTACH_POSITION); ac++;
      XtSetArg(al[ac],XmNbottomPosition, 72+y*9); ac++;
      command[y]=XmCreatePushButton(form,"label",al,ac);
      XtManageChild(command[y]);
      XtAddCallback(command[y],XmNactivateCallback,
        commandCB,(int *) cmd_number);
    }
  }
  
  
  /* Create prompt dialogs and leave them unmanaged until needed */
  
  /* Target azimuth */

  ac=0;
  XtSetArg(al[ac], XmNselectionLabelString,
    XmStringCreateLtoR("Enter azimuth (deg): ",char_set)); ac++;
  XtSetArg(al[ac], XmNdialogTitle,
    XmStringCreateLtoR("Target",char_set)); ac++;
  target_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
  XtAddCallback(target_dialog, XmNokCallback, target_dialogCB, (int *) OK);
  XtAddCallback(target_dialog, XmNcancelCallback, target_dialogCB, (int *) CANCEL);
  XtUnmanageChild (XmSelectionBoxGetChild (target_dialog, XmDIALOG_HELP_BUTTON));

  /* Create the menubar */

  create_menus(menu_bar); 

  /* Service status display and contact with dome */

  poll_interval_id = XtAppAddTimeOut(context,
    poll_interval,poll_interval_handler,poll_interval_data_ptr);  
  
  /* Start the user interface */

  XtRealizeWidget(toplevel);
  XtAppMainLoop(context);

  exit(0);
}


/* Add an item into a menu in the menubar */   

Widget make_menu_item(item_name,client_data,menu)
  char *item_name;
  int client_data;
  Widget menu;
{
  int ac;
  Arg al[10];
  Widget item;

  ac = 0;
  XtSetArg(al[ac], XmNlabelString,
    XmStringCreateLtoR(item_name,char_set)); ac++;      
  item=XmCreatePushButton(menu,item_name,al,ac);
  XtManageChild(item);
  XtAddCallback (item,XmNactivateCallback,menuCB,(int *) client_data);      
  XtSetSensitive(item,True);
  return(item);
}


/* Create a menu on the menu bar */

Widget make_menu(menu_name,menu_bar)
  char *menu_name; 
  Widget menu_bar;
{
  int ac;
  Arg al[10];
  Widget menu, cascade;

  menu=XmCreatePulldownMenu(menu_bar,menu_name,NULL,0);
  ac=0;
  XtSetArg(al[ac],XmNsubMenuId,menu);  ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR(menu_name,char_set)); ac++;
  cascade=XmCreateCascadeButton(menu_bar,menu_name,al,ac);      
  XtManageChild (cascade); 
  return(menu);
}


/* Create all the menubar entries organized into submenus*/

void create_menus(Widget menu_bar)
{
     
  /* Create the file menu */

  file_menu = make_menu("File",menu_bar);
  file_exit_item = make_menu_item("Exit",FILEEXIT,file_menu);
  

  /* Create the set menu */

  set_menu = make_menu("Set",menu_bar);


  set_target_item = make_menu_item("Target",SETTARGETAZ,set_menu);
  
}

/* Callback on menubar action sorts requests based on client_data */

void menuCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{

  if (client_data==FILEEXIT)   
  {
    
    /* Make sure everything is turned off at last contact */
  
    domecmd = DOMEOFF;
    ControlDome(domecmd);

    exit(0);
  }
  
  if (client_data==SETTARGETAZ)
  {
    XtManageChild(target_dialog);
  }
  
  /* Take no action on cancel */
  
  if (client_data==CANCEL); 

}

/* Dialog for user input of target azimuth */

void target_dialogCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmSelectionBoxCallbackStruct  *call_data;
{    
  char *s;
  double newtargetaz;

  switch (client_data)
  {
    case OK:
      XmStringGetLtoR(call_data->value,char_set,&s);
      newtargetaz = atof(s);
      if ( ( newtargetaz >= lowerlimit ) && ( newtargetaz <= upperlimit ) )
      {
        dometargetaz = newtargetaz;
      }
      else
      {
        dometargetaz = domeaz;
      }
      XtFree(s);
      break;
    case CANCEL:
      break;
  }
  update_target_display();
  XtUnmanageChild(w);
}


/* Callback function for the radio buttons that select auto function */

void toggleCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
{
    int n;

    /* Identify the radio button */
    n=client_data;
       
    /* Assign functionality to the panel guide pushbuttons */

    switch (n) 
    {
       /* Pushbutton turns on manual control */    
       case 0: 
         panelautotrack = FALSE;
         break; 
          
       /* Pushbutton turns on auto track */
       case 1: 
         panelautotrack = TRUE;
         break;                   
    } 
}

/* Callback function for the 9 grid buttons */
/* Top to bottom, then left to right */

void buttonCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n;
  int et = call_event->type;
  int dn,up;
   
  /* identify the button number n = 3*x+y */

  n=client_data;

  /* identify the action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;
      
  /* respond to the request */
  
  if(dn)
  {
    switch (n) 
    {
   
      /* NoOp*/
      case 0:
      break;

      /* CW */
      case 1: 
        domecmd = CWROTATION;
        ControlDome(domecmd);
      break;

      /* NoOp */
      case 2: 
      break;

      /* NoOp */
      case 3: 
      break;

      /* Stop */
      case 4:
        domecmd = STOPROTATION;
        ControlDome(domecmd);
       break;

      /* NoOp */
      case 5: 
      break;

      /* NoOp */
      case 6:
      break;

      /* CCW */
      case 7:
        domecmd = CCWROTATION;
        ControlDome(domecmd);
      break;

      /* NoOp */
      case 8: 
      break;    

    }
        
  }
  
  if(up)
  {     
    /* Wait for the dome to respond before accepting another cycle */

    usleep(1000000);
  }
}


/* Callback for the dome command buttons */
/* Buttons numbered from 0, top to bottom, then left to right */

void commandCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  
  int n, cmd_number, nloops;
  double telra, teldec;

  /* Identify the button */
  cmd_number = client_data;

  /* Respond to the request */

  n = cmd_number;

  switch (n) 
  {

    /* Set telescope target */  
    case 0:
      dometargetaz = 180.;
      nloops = 4;
      GetTel(&telra, &teldec);
      SolveDomeAzimuth(telra, teldec, &dometargetaz, nloops);
      break; 
       
    /* Slew to target */
    case 1:
      domegoto = TRUE;
      break;

    /* Stop dome operations */
    case 2:
      domecmd = DOMEOFF;
      ControlDome(domecmd);
      domegoto = FALSE;
      break;  
      
    /* Set north target */  
    case 3:
      dometargetaz = 0.;
      break;  
    
    /* Set south target */
    case 4:
      dometargetaz = 180.;
      break; 
      
    /* Set flat field target */  
    case 5:
      dometargetaz = 90.;
      break;  
      
     /* Open upper shutter */
    case 6:
      domecmd = OPENUPPER;
      ControlDome(domecmd);
      break;
    
    /* Close upper shutter */
    case 7:
      domecmd = CLOSEUPPER;
      ControlDome(domecmd);
      break; 
      
    /* Stop upper shutter */
    case 8:
      domecmd = STOPUPPER;
      ControlDome(domecmd);
      break;  

    /* Open lower shutter */  
    case 9:
      domecmd =  OPENLOWER;
      ControlDome(domecmd);
      break;             
    
    /* Close lower shutter */
    case 10:
      domecmd = CLOSELOWER;
      ControlDome(domecmd);
      break;  
            
    /* Stop lower shutter */  
    case 11:
      domecmd = STOPLOWER;
      ControlDome(domecmd);
      break; 

  }
  update_all_displays();
}


/* Update current azimuth in the window */

void update_azimuth_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;
  
  sprintf(buf,"%4.1lf",domeaz);

  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[0],al,ac);   
  XmStringFree(xmstr);
}    

/* Update azimuth target in the window */

void update_target_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;

  sprintf(buf,"%4.1lf",dometargetaz);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[1],al,ac);   
  XmStringFree(xmstr);
}  

/* Update CW rotation activity in the window */

void update_cw_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;

  if ( (domestatus & ROTATINGCW) == ROTATINGCW ) 
  {
    sprintf(buf,"        On        ");
  }
  else
  {
    sprintf(buf,"        Off       ");
  }
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[2],al,ac);   
  XmStringFree(xmstr);
} 

/* Update CCW rotation activity in the window */

void update_ccw_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;

  if ( (domestatus & ROTATINGCCW) == ROTATINGCCW )
  {
    sprintf(buf,"        On        ");
  }
  else 
  {
    sprintf(buf,"        Off       ");
  }
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[3],al,ac);   
  XmStringFree(xmstr);
}

/* Update upper shutter state in the window */

void update_upper_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;

  if ( (upper_shutter_status & NOTCLOSED) == NOTCLOSED )
  {
    sprintf(buf,"     Open         ");
  }
  else if ( (upper_shutter_status & OPENLIMIT) == OPENLIMIT ) 
  {
    sprintf(buf,"   Fully Open     ");    
  }
  else if ( (upper_shutter_status & OPENING) == OPENING ) 
  {
    sprintf(buf,"     Opening     ");    
  }  
  else if ( (upper_shutter_status & CLOSING) == CLOSING ) 
  {
    sprintf(buf,"     Closing     ");    
  }    
  else  
  {
    sprintf(buf,"     Closed       ");    
  }
  
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[4],al,ac);   
  XmStringFree(xmstr);
}     

/* Update lower shutter state in the window */

void update_lower_display()
{
  XmString xmstr;
  Arg al[10];
  int ac;

  if ( (lower_shutter_status & NOTCLOSED) == NOTCLOSED )
  {
    sprintf(buf,"     Open         ");
  }
  else if ( (lower_shutter_status & OPENLIMIT) == OPENLIMIT ) 
  {
    sprintf(buf,"   Fully Open     ");    
  }
  else if ( (lower_shutter_status & OPENING) == OPENING ) 
  {
    sprintf(buf,"     Opening     ");    
  }  
  else if ( (lower_shutter_status & CLOSING) == CLOSING ) 
  {
    sprintf(buf,"     Closing     ");    
  }    
  else  
  {
    sprintf(buf,"     Closed       ");    
  }
  
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(show_info[5],al,ac);   
  XmStringFree(xmstr);
}    

/* Update all display buffers */

void update_all_displays()
{
  update_azimuth_display();
  update_target_display(); 
  update_cw_display();
  update_ccw_display();
  update_upper_display();
  update_lower_display();
}   


/* Manage the polling */

void poll_interval_handler(XtPointer client_data_ptr, XtIntervalId *client_id) 
{
    
  /* Manage the dome */
  
  dome();  
            
  /* Start the timer again.  It will return to the handler after poll_interval. */

  poll_interval_id = XtAppAddTimeOut(context, poll_interval,
    poll_interval_handler, poll_interval_data_ptr);
}  


void dome(void)
{
  static int tcount;
  int nloops;
  double telra, teldec;  
  
  /* Local very high priority operations would go here  */

  if (domegoto == TRUE)
  {
    domegoto = DomeToAzimuth(dometargetaz);
  }

  /* Otherwise poll dome every 16 seconds for a 250 ms polling cycle */
  
  if ((tcount < 0) || (tcount > 64))
  {
    tcount = 1;  
    
    /* Get the latest dome azimuth and state */
    
    GetDome(&domeaz,&domestatus); 
    GetShutter(&upper_shutter_status,&lower_shutter_status); 
    
    if (panelautotrack == TRUE)
    {
      dometargetaz = 180.;
      nloops = 4;
      GetTel(&telra, &teldec);
      SolveDomeAzimuth(telra, teldec, &dometargetaz, nloops);
      if ( (dometargetaz >= lowerlimit) && (dometargetaz <= upperlimit) )
      {        
        domegoto = DomeToAzimuth(dometargetaz);
      }  
    }

    /* Update the panel */
    
    update_all_displays();
  }
  tcount++; 

}

/* Contact the dome */

void initdome(void)
{
  
  static int inited;
  
  if (inited)
  { 
    return;
  }
  
  /* Mounting and dome defaults */
  
  telmount = GEM;
  dec_offset = DECOFFSET;
  dec_height = DECHEIGHT;
  dec_length = DECLENGTH;
  dome_radius = DOMERADIUS;
  
  /* Longitude and latitude defaults */
  
  site_longitude = LONGITUDE;
  site_latitude = LATITUDE;
  
  /* For compatibility with previous versions of algorithms.c */
  
  SiteLongitude = site_longitude;
  SiteLatitude = site_latitude;  
  
  /* Handle the configuration file here if needed */
      
  fprintf(stderr,"Connecting to the dome ... \n");
  
  /* Read configuration file to set necessary parameters */
  
  read_config();
  
  /* Set dome limits */
      
  if ( upperlimit > lowerlimit )
  {
    SetDomeLimits(lowerlimit, upperlimit);
    fprintf(stderr,
      "Warning: dome azimuth range limited from %4.1lf to %4.1lf degrees.\n", 
      lowerlimit, upperlimit);     
  }
   
  /* Make sure everything is turned off at first contact */
  
  domecmd = DOMEOFF;
  ControlDome(domecmd);
  
  /* Set flag so we do the initialization only once */

  inited = 1;

}

/* Read and parse the initial configuration file */

void read_config(void)
{
  char configstr[121];
  char *configptr = configstr;
  char *configfile;  
  FILE *fp_config; 
  
  configfile = (char *) malloc (MAXPATHLEN);
  strcpy(configfile,CONFIGFILE);  
  fp_config = fopen(configfile, "r");
  
  if ( fp_config == NULL )
  {
    fprintf(stderr,"Caution: required dome configuration not found.\n");
    return;
  }
  
  while ( configstr == fgets(configstr,80,fp_config) )
  {
    
    configptr = strstr(configstr,"dome.longitude");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &site_longitude);
        SiteLongitude = site_longitude;
      }  
    }
    configptr = strstr(configstr,"dome.latitude");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &site_latitude);
        SiteLatitude = site_latitude;
      }  
    }
    configptr = strstr(configstr,"dome.mount");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%d", &telmount);
      }  
    }
    configptr = strstr(configstr,"dome.decoffset");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_offset);
      }  
    }
    configptr = strstr(configstr,"dome.decheight");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_height);
      }  
    }
    configptr = strstr(configstr,"dome.declength");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dec_length);
      }  
    }
    configptr = strstr(configstr,"dome.radius");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &dome_radius);
      }  
    }
    configptr = strstr(configstr,"dome.upperlimit");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &upperlimit);
      }  
    }
    configptr = strstr(configstr,"dome.lowerlimit");
    if ( configptr != NULL)
    {
      configptr = strstr(configstr,"=");
      if ( configptr != NULL)
      {
        configptr = configptr + 1;
        sscanf(configptr,"%lf", &lowerlimit);
      }  
    }

  }
     
  fclose(fp_config);
}     
      
